"""Experimental and undocumented. Speak to Max."""

from .restraint import Restraint


def write_amber_restraints(
    restraint_list, filename_prefix, zfill=2, filename_suffix=".RST"
):
    import mcol
    import mout
    import os

    os.system("mkdir -p " + os.path.dirname(filename_prefix))

    assert isinstance(restraint_list, list)

    # get all the restraint values
    all_values = []
    for res in restraint_list:
        assert isinstance(res, Restraint)
        values = res.values()
        assert values is not None
        all_values.append(values)
        assert len(values) == len(all_values[0])

    num_windows = len(all_values[0])

    for i in range(num_windows):

        i_str = str(i).zfill(zfill)

        filename = filename_prefix + i_str + filename_suffix

        # print(i)
        # print(filename)

        rst_buffer = "# " + filename + "\n"

        for restraint in restraint_list:
            rst_buffer += restraint.amber_block(i)

        # print(rst_buffer)

        out_rst = open(filename, "w")
        out_rst.write(rst_buffer)
        out_rst.close()
        mout.out("File written to " + mcol.file + filename)

    import pickle

    pickle_file = os.path.dirname(filename_prefix) + "/amp_res.pkl"
    with open(pickle_file, "wb") as file:
        pickle.dump(restraint_list, file)
    mout.out("Restraints dumped to to " + mcol.file + pickle_file)


# exit()


def umbrella_plotter(filenames, bins=20, subdir=None, show_level=2):
    import os
    import mplot

    from . import signal

    big_ydata = []
    labels = []

    if subdir is not None:
        os.system("mkdir -p " + subdir)

    for file in filenames:

        print(file)

        xdata, ydata = signal.parseDat(file, num_columns=2, pre_strip=True)

        label = os.path.basename(file).replace(".rco", "")
        labels.append(label)

        plotfile = None
        if subdir is not None:
            plotfile = subdir + "/" + label + ".png"

        if show_level > 1:
            show = True
        else:
            show = False

        mplot.hist1D(
            ydata,
            show=show,
            xlab="Reaction Coordinate",
            ylab="Frequency",
            bins=bins,
            title=label,
            filename=plotfile,
        )

        big_ydata.append(ydata)

    if show_level > 0:
        show = True
    else:
        show = False

    plotfile = None
    if subdir is not None:
        plotfile = subdir + "/allwindows.png"

    mplot.graph2D(
        xdata,
        big_ydata,
        show=show,
        filename=plotfile,
        ytitles=labels,
        xlab="MD Step",
        ylab="Reaction Coordinate",
    )


def umbrella_helper_2dist(
    atoms,
    weights,
    coord_range,
    num_windows,
    force_constant,
    harmonic_width,
    subdir=None,
    samples=1000,
    graph=False,
):
    import mcol
    import mout
    import os
    import mplot

    assert len(atoms) == 4
    assert len(weights) == 2

    mout.headerOut("Atoms")

    for i, atom in enumerate(atoms):
        mout.varOut(
            "Atom #" + str(i + 1),
            [atom.name, atom.residue, atom.pdb_index],
            integer=True,
            list_length=False,
        )

    mout.headerOut("Windows")
    mout.varOut("#Windows", num_windows, valCol=mcol.arg)

    centres = []

    for i in range(num_windows):
        centres.append(
            coord_range[0] + i * (coord_range[1] - coord_range[0]) / (num_windows - 1)
        )

        mout.varOut("Window #" + str(i + 1) + " Centre", centres[-1], precision=4)

    mout.headerOut("Amber Restraint File:")

    if subdir is not None:
        os.system("mkdir -p " + subdir)

    restraints = []

    for i, centre in enumerate(centres):

        end = "\n"

        rst_buffer = "# window_" + str(i + 1) + ".RST" + end
        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[0].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[2].pdb_index) + ","
        rst_buffer += str(atoms[3].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[0]) + ","
        rst_buffer += str(weights[1]) + "," + end

        r2 = centre
        r3 = centre

        r1 = centre - harmonic_width
        r4 = centre + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end + end

        restraints.append(
            {
                "centre": centre,
                "r1": r1,
                "r2": r2,
                "r3": r3,
                "r4": r4,
                "rk2": rk2,
                "rk3": rk3,
            }
        )

        if subdir is not None:
            out_rst = open(subdir + "/window_" + str(i + 1).zfill(2) + ".RST", "w")
            out_rst.write(rst_buffer)
            out_rst.close()
            mout.out(
                "File written to "
                + mcol.file
                + subdir
                + "/window_"
                + str(i + 1).zfill(2)
                + ".RST"
            )

    xdata = []
    big_ydata = []

    for i in range(samples):
        # print(i,coord_range[0]+i*(coord_range[1]-coord_range[0])/(samples-1))
        x = coord_range[0] + (i * 1.4 / (samples - 1) - 0.2) * (
            coord_range[1] - coord_range[0]
        )
        xdata.append(x)

    for restraint in restraints:

        ydata = []

        r1 = restraint["r1"]
        r2 = restraint["r2"]
        r3 = restraint["r3"]
        r4 = restraint["r4"]
        rk2 = restraint["rk2"]
        rk3 = restraint["rk3"]

        for x in xdata:
            y = restraint_potential(x, r1, r2, r3, r4, rk2, rk3)

            ydata.append(y)

        big_ydata.append(ydata)

    # if graph:
    if subdir is not None:
        graphfile = subdir + "/allwindows.png"
    else:
        graphfile = None

    if subdir is not None and graph is not None:
        mplot.graph2D(
            xdata, big_ydata, show=graph, ymax=120, ymin=-5, filename=graphfile
        )

    if subdir is not None:

        pot_buffer = "# restraint potentials" + end

        pot_buffer += str(x) + " "

        for restraint in restraints:
            r1 = restraint["r1"]
            r2 = restraint["r2"]
            r3 = restraint["r3"]
            r4 = restraint["r4"]
            rk2 = restraint["rk2"]
            rk3 = restraint["rk3"]
            y = restraint_potential(x, r1, r2, r3, r4, rk2, rk3)
            pot_buffer += str(y) + " "

        out_dat = open(subdir + "/allwindows.dat", "w")
        out_dat.write(pot_buffer)
        out_dat.close()


def umb_rst_2prot(
    atoms,
    weights,
    coord_range,
    num_windows,
    force_constant,
    harmonic_width,
    subdir=None,
    samples=1000,
    graph=False,
    adiab_windows=True,
    fix_angle=None,
    fix_third=None,
):
    import mcol
    import mout
    import os
    import mplot

    assert len(atoms) == 6
    assert len(weights) == 2

    # 2x (donor-hydrogen...acceptor)

    mout.headerOut("Atoms")

    for i, atom in enumerate(atoms):
        mout.varOut(
            "Atom #" + str(i + 1),
            [atom.name, atom.residue, atom.pdb_index],
            integer=True,
            list_length=False,
        )

    mout.headerOut("Windows")
    mout.varOut("#Windows", num_windows, valCol=mcol.arg)

    centres = []

    for i in range(num_windows):
        centres.append(
            coord_range[0] + i * (coord_range[1] - coord_range[0]) / (num_windows - 1)
        )

        mout.varOut("Window #" + str(i + 1) + " Centre", centres[-1], precision=4)

    mout.headerOut("Amber Restraint File:")

    if subdir is not None:
        os.system("mkdir -p " + subdir)

    restraints = []

    ### RC:
    ### 		dist(donor-hydrogen)	  [1.0->2.0]
    ### 	  - dist(hydrogen...acceptor) [2.0->1.0]
    ### 	  =			   				  [-1.0->1.0]

    mout.array2file("centres.dat", centres)

    if fix_angle is not None:
        mout.warningOut("Angle restraints are active!")
    if fix_third is not None:
        mout.warningOut("Third bond restraints are active!")

    for i, centre in enumerate(centres):

        end = "\n"

        rst_buffer = "# window_" + str(i + 1) + ".RST" + end

        # First reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[0].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[2].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[0]) + ","
        rst_buffer += str(-weights[0]) + "," + end

        r2 = centre
        r3 = centre

        r1 = centre - harmonic_width
        r4 = centre + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        # Second reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[3].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[5].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[1]) + ","
        rst_buffer += str(-weights[1]) + "," + end

        r2 = centre
        r3 = centre

        r1 = centre - harmonic_width
        r4 = centre + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        restraints.append(
            {
                "centre": centre,
                "r1": r1,
                "r2": r2,
                "r3": r3,
                "r4": r4,
                "rk2": rk2,
                "rk3": rk3,
            }
        )

        ###### ANGLE FIX ######

        # fix_angle should be a list with [atom,atom,atom,angle]

        if fix_angle is not None:
            # try to catch some errors
            assert isinstance(fix_angle, list)
            assert len(fix_angle) == 4

            # construct the restraint

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(fix_angle[0].pdb_index) + ","
            rst_buffer += str(fix_angle[1].pdb_index) + ","
            rst_buffer += str(fix_angle[2].pdb_index) + "," + end

            r2 = fix_angle[-1]
            r3 = 180

            r1 = r2 - 20
            r4 = r2 + 20

            rk2 = force_constant
            rk3 = force_constant

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

        ###### THIRD BOND FIX ######

        # fix_third should be a list with [atom,atom,atom]
        # [donor,H,acceptor]

        if fix_third is not None:
            # try to catch some errors
            assert isinstance(fix_third, list)
            assert len(fix_third) == 3

            # construct the restraints

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(fix_third[0].pdb_index) + ","
            rst_buffer += str(fix_third[1].pdb_index) + "," + end

            r2 = 0.9
            r3 = 1.5

            r1 = 0.0
            r4 = 2.4

            rk2 = force_constant
            rk3 = force_constant

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(fix_third[1].pdb_index) + ","
            rst_buffer += str(fix_third[2].pdb_index) + "," + end

            r2 = 1.4
            r3 = 2.4

            r1 = 0.5
            r4 = 3.3

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

        rst_buffer += end

        if subdir is not None:
            out_rst = open(subdir + "/window_" + str(i + 1).zfill(2) + ".RST", "w")
            out_rst.write(rst_buffer)
            out_rst.close()
            mout.out(
                "File written to "
                + mcol.file
                + subdir
                + "/window_"
                + str(i + 1).zfill(2)
                + ".RST"
            )

    if adiab_windows:

        for i, centre in enumerate(centres):

            end = "\n"

            rst_buffer = "# adiab_" + str(i + 1) + ".RST" + end

            # First reaction coord

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(atoms[0].pdb_index) + ","
            rst_buffer += str(atoms[1].pdb_index) + ","
            rst_buffer += str(atoms[1].pdb_index) + ","
            rst_buffer += str(atoms[2].pdb_index) + "," + end

            rst_buffer += "rstwt="
            rst_buffer += str(weights[0]) + ","
            rst_buffer += str(-weights[0]) + "," + end

            r2 = centre
            r3 = centre

            r1 = centre - harmonic_width * 10
            r4 = centre + harmonic_width * 10

            rk2 = force_constant * 50
            rk3 = force_constant * 50

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

            # Second reaction coord

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(atoms[3].pdb_index) + ","
            rst_buffer += str(atoms[4].pdb_index) + ","
            rst_buffer += str(atoms[4].pdb_index) + ","
            rst_buffer += str(atoms[5].pdb_index) + "," + end

            rst_buffer += "rstwt="
            rst_buffer += str(weights[1]) + ","
            rst_buffer += str(-weights[1]) + "," + end

            r2 = centre
            r3 = centre

            r1 = centre - harmonic_width * 10
            r4 = centre + harmonic_width * 10

            rk2 = force_constant * 50
            rk3 = force_constant * 50

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end + end

            if subdir is not None:
                out_rst = open(subdir + "/adiab_" + str(i + 1).zfill(2) + ".RST", "w")
                out_rst.write(rst_buffer)
                out_rst.close()
                mout.out(
                    "File written to "
                    + mcol.file
                    + subdir
                    + "/adiab_"
                    + str(i + 1).zfill(2)
                    + ".RST"
                )

    xdata = []
    big_ydata = []

    for i in range(samples):
        # print(i,coord_range[0]+i*(coord_range[1]-coord_range[0])/(samples-1))
        x = coord_range[0] + (i * 1.4 / (samples - 1) - 0.2) * (
            coord_range[1] - coord_range[0]
        )
        xdata.append(x)

    for restraint in restraints:

        ydata = []

        r1 = restraint["r1"]
        r2 = restraint["r2"]
        r3 = restraint["r3"]
        r4 = restraint["r4"]
        rk2 = restraint["rk2"]
        rk3 = restraint["rk3"]

        for x in xdata:
            y = restraint_potential(x, r1, r2, r3, r4, rk2, rk3)

            ydata.append(y)

        big_ydata.append(ydata)

    # if graph:
    if subdir is not None:
        graphfile = subdir + "/allwindows.png"
    else:
        graphfile = None

    if subdir is not None and graph is not None:
        mplot.graph2D(
            xdata, big_ydata, show=graph, ymax=120, ymin=-5, filename=graphfile
        )

    if subdir is not None:

        pot_buffer = "# restraint potentials" + end

        pot_buffer += str(x) + " "

        for restraint in restraints:
            r1 = restraint["r1"]
            r2 = restraint["r2"]
            r3 = restraint["r3"]
            r4 = restraint["r4"]
            rk2 = restraint["rk2"]
            rk3 = restraint["rk3"]
            y = restraint_potential(x, r1, r2, r3, r4, rk2, rk3)
            pot_buffer += str(y) + " "

        out_dat = open(subdir + "/allwindows.dat", "w")
        out_dat.write(pot_buffer)
        out_dat.close()


def umb_rst_2prot_new(
    atoms,
    weights,
    coord_range,
    num_windows,
    force_constant,
    harmonic_width,
    subdir=None,
    samples=1000,
    graph=False,
    adiab_windows=True,
    fix_angle=None,
    fix_third=None,
    add_len=None,
):
    import mcol
    import mout
    import os

    assert len(atoms) == 6
    assert len(weights) == 2

    # 2x (donor-hydrogen...acceptor)

    mout.headerOut("Atoms")

    for i, atom in enumerate(atoms):
        mout.varOut(
            "Atom #" + str(i + 1),
            [atom.name, atom.residue, atom.pdb_index],
            integer=True,
            list_length=False,
        )

    mout.headerOut("Windows")
    mout.varOut("#Windows", num_windows, valCol=mcol.arg)

    print(coord_range)

    if add_len is not None:
        coord_range[0] -= 0.5 * add_len
        coord_range[1] += 0.5 * add_len
        coord_range[2] -= 0.5 * add_len
        coord_range[3] += 0.5 * add_len
    print(coord_range)

    centres = []

    for i in range(num_windows):
        centre_1 = coord_range[0] + i * (coord_range[1] - coord_range[0]) / (
            num_windows - 1
        )
        centre_2 = coord_range[2] + i * (coord_range[3] - coord_range[2]) / (
            num_windows - 1
        )

        centres.append([centre_1, centre_2])

        mout.varOut("Window #" + str(i + 1) + " Centres", centres[-1], precision=4)

    mout.headerOut("Amber Restraint File:")

    if subdir is not None:
        os.system("mkdir -p " + subdir)

    restraints = []

    ### RC:
    ### 		dist(donor-hydrogen)	  [1.0->2.0]
    ### 	  - dist(hydrogen...acceptor) [2.0->1.0]
    ### 	  =			   				  [-1.0->1.0]

    mout.array2file("centres.dat", centres)

    if fix_angle is not None:
        mout.warningOut("Angle restraints are active!")
    if fix_third is not None:
        mout.warningOut("Third bond restraints are active!")

    for i, centre in enumerate(centres):

        end = "\n"

        rst_buffer = "# window_" + str(i + 1) + ".RST" + end

        # First reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[0].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[2].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[0]) + ","
        rst_buffer += str(-weights[0]) + "," + end

        r2 = centre[0]
        r3 = centre[0]

        r1 = centre[0] - harmonic_width
        r4 = centre[0] + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        # Second reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[3].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[5].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[1]) + ","
        rst_buffer += str(-weights[1]) + "," + end

        r2 = centre[1]
        r3 = centre[1]

        r1 = centre[1] - harmonic_width
        r4 = centre[1] + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        # restraints.append({'centre':centre,
        # 				   'r1':r1,
        # 				   'r2':r2,
        # 				   'r3':r3,
        # 				   'r4':r4,
        # 				   'rk2':rk2,
        # 				   'rk3':rk3})

        ###### ANGLE FIX ######

        # fix_angle should be a list with [atom,atom,atom,angle]

        if fix_angle is not None:
            # try to catch some errors
            assert isinstance(fix_angle, list)
            assert len(fix_angle) == 4

            # construct the restraint

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(fix_angle[0].pdb_index) + ","
            rst_buffer += str(fix_angle[1].pdb_index) + ","
            rst_buffer += str(fix_angle[2].pdb_index) + "," + end

            r2 = fix_angle[-1]
            r3 = 180

            r1 = r2 - 20
            r4 = r2 + 20

            rk2 = force_constant
            rk3 = force_constant

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

        ###### THIRD BOND FIX ######

        # fix_third should be a list with [atom,atom,atom]
        # [donor,H,acceptor]

        if fix_third is not None:

            # try to catch some errors
            assert isinstance(fix_third, list)
            assert len(fix_third) == 3

            # construct the restraints

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(fix_third[0].pdb_index) + ","
            rst_buffer += str(fix_third[1].pdb_index) + "," + end

            if add_len is not None:
                assert isinstance(add_len, float) or isinstance(add_len, int)
                this_add = add_len
            else:
                this_add = 0.0

            r2 = 0.9
            r3 = 1.5

            r1 = 0.0
            r4 = 2.4

            rk2 = force_constant
            rk3 = force_constant

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(fix_third[1].pdb_index) + ","
            rst_buffer += str(fix_third[2].pdb_index) + "," + end

            r2 = 1.4 + this_add
            r3 = 2.4 + this_add

            r1 = 0.5 + this_add
            r4 = 3.3 + this_add

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

        if add_len is not None:
            # Additional restraints to extend hydrogen bonds

            assert isinstance(add_len, float) or isinstance(add_len, int)

            # construct the restraints

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(atoms[0].pdb_index) + ","
            rst_buffer += str(atoms[2].pdb_index) + "," + end

            r2 = 2.5 + add_len
            r3 = 3.0 + add_len

            r1 = r2 - 1.0
            r4 = r3 + 1.0

            rk2 = force_constant
            rk3 = force_constant

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(atoms[3].pdb_index) + ","
            rst_buffer += str(atoms[5].pdb_index) + "," + end

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

        rst_buffer += end

        if subdir is not None:
            out_rst = open(subdir + "/window_" + str(i + 1).zfill(2) + ".RST", "w")
            out_rst.write(rst_buffer)
            out_rst.close()
            mout.out(
                "File written to "
                + mcol.file
                + subdir
                + "/window_"
                + str(i + 1).zfill(2)
                + ".RST"
            )


def umb_rst_2prot_1rc(
    atoms,
    weights,
    coord_range,
    num_windows,
    force_constant,
    harmonic_width,
    subdir=None,
    samples=1000,
    graph=False,
    adiab_windows=False,
    fix_length=False,
):
    import mcol
    import mout
    import os

    assert len(atoms) == 6
    assert len(weights) == 2

    # 2x (donor-hydrogen...acceptor)

    mout.headerOut("Atoms")

    for i, atom in enumerate(atoms):
        mout.varOut(
            "Atom #" + str(i + 1),
            [atom.name, atom.residue, atom.pdb_index],
            integer=True,
            list_length=False,
        )

    mout.headerOut("Windows")
    mout.varOut("#Windows", num_windows, valCol=mcol.arg)

    centres = []

    for i in range(num_windows):
        centres.append(
            coord_range[0] + i * (coord_range[1] - coord_range[0]) / (num_windows - 1)
        )

        mout.varOut("Window #" + str(i + 1) + " Centre", centres[-1], precision=4)

    mout.headerOut("Amber Restraint File:")

    if subdir is not None:
        os.system("mkdir -p " + subdir)

    restraints = []

    ### RC:
    ### 		dist(donor-hydrogen)	  [1.0->2.0]
    ### 	  - dist(hydrogen...acceptor) [2.0->1.0]
    ### 	  =			   				  [-1.0->1.0]

    for i, centre in enumerate(centres):

        end = "\n"

        rst_buffer = "# window_" + str(i + 1) + ".RST" + end

        # First reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[0].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[2].pdb_index) + ","
        rst_buffer += str(atoms[3].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[5].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[0]) + ","
        rst_buffer += str(-weights[0]) + ","
        rst_buffer += str(weights[1]) + ","
        rst_buffer += str(-weights[1]) + "," + end

        r2 = centre
        r3 = centre

        r1 = centre - harmonic_width
        r4 = centre + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        # Output only dummy-restraints

        rk2 = 0.0
        rk3 = 0.0

        rst_buffer += "&rst" + end
        rst_buffer += "iat="
        rst_buffer += str(atoms[0].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[2].pdb_index) + "," + end
        rst_buffer += "rstwt="
        rst_buffer += str(weights[0]) + ","
        rst_buffer += str(-weights[0]) + "," + end
        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end
        rst_buffer += "/" + end

        # Second reaction coord

        rst_buffer += "&rst" + end
        rst_buffer += "iat="
        rst_buffer += str(atoms[3].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[5].pdb_index) + "," + end
        rst_buffer += "rstwt="
        rst_buffer += str(weights[1]) + ","
        rst_buffer += str(-weights[1]) + "," + end
        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end
        rst_buffer += "/" + end

        # Restrain donor-acceptor H-bond lengths

        if fix_length:
            r1 = 0.0
            r2 = 2.0
            r3 = 4.0
            r4 = 6.0

            rk2 = force_constant * 2
            rk3 = force_constant * 2

            rst_buffer += "&rst" + end
            rst_buffer += "iat="
            rst_buffer += str(atoms[0].pdb_index) + ","
            rst_buffer += str(atoms[2].pdb_index) + "," + end
            rst_buffer += "rstwt="
            rst_buffer += str(weights[1]) + "," + end
            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end
            rst_buffer += "/" + end

            rst_buffer += "&rst" + end
            rst_buffer += "iat="
            rst_buffer += str(atoms[3].pdb_index) + ","
            rst_buffer += str(atoms[5].pdb_index) + "," + end
            rst_buffer += "rstwt="
            rst_buffer += str(weights[1]) + "," + end
            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end
            rst_buffer += "/" + end

        rst_buffer += end

        if subdir is not None:
            out_rst = open(subdir + "/window_" + str(i + 1).zfill(2) + ".RST", "w")
            out_rst.write(rst_buffer)
            out_rst.close()
            mout.out(
                "File written to "
                + mcol.file
                + subdir
                + "/window_"
                + str(i + 1).zfill(2)
                + ".RST"
            )

    if adiab_windows:
        mout.warningOut("Adiabatic windows not implemented")


def umb_rst_2prot_asy(
    atoms,
    weights,
    coord_range,
    num_windows,
    force_constant,
    harmonic_width,
    subdir=None,
    samples=1000,
    graph=False,
    adiab_windows=True,
    reverse=False,
):
    import mcol
    import mout
    import os

    num_windows = num_windows // 2 + 1

    assert len(atoms) == 6
    assert len(weights) == 2

    # 2x (donor-hydrogen...acceptor)

    mout.headerOut("Atoms")

    for i, atom in enumerate(atoms):
        mout.varOut(
            "Atom #" + str(i + 1),
            [atom.name, atom.residue, atom.pdb_index],
            integer=True,
            list_length=False,
        )

    mout.headerOut("Windows")
    mout.varOut("#Windows", num_windows, valCol=mcol.arg)

    centres = []

    for i in range(num_windows):
        centres.append(
            coord_range[0] + i * (coord_range[1] - coord_range[0]) / (num_windows - 1)
        )

        mout.varOut("Window #" + str(i + 1) + " Centre", centres[-1], precision=4)

    centres_2d = []

    if not reverse:
        for centre in centres:
            centres_2d.append([centre, centres[0]])
        for centre in centres:
            if centre == centres[0]:
                continue
            centres_2d.append([centres[-1], centre])
    else:
        for centre in centres:
            centres_2d.append([centres[0], centre])
        for centre in centres:
            if centre == centres[0]:
                continue
            centres_2d.append([centre, centres[-1]])

    mout.headerOut("Amber Restraint File:")

    # mout.varOut("centres",centres_2d,precision=2,sf=False)

    if subdir is not None:
        os.system("mkdir -p " + subdir)

    restraints = []

    ### RC:
    ### 		dist(donor-hydrogen)	  [1.0->2.0]
    ### 	  - dist(hydrogen...acceptor) [2.0->1.0]
    ### 	  =			   				  [-1.0->1.0]

    for i, centre_pair in enumerate(centres_2d):

        end = "\n"

        rst_buffer = "# window_" + str(i + 1) + ".RST" + end

        # First reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[0].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[1].pdb_index) + ","
        rst_buffer += str(atoms[2].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[0]) + ","
        rst_buffer += str(-weights[0]) + "," + end

        r2 = centre_pair[0]
        r3 = centre_pair[0]

        r1 = centre_pair[0] - harmonic_width
        r4 = centre_pair[0] + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        # Second reaction coord

        rst_buffer += "&rst" + end

        rst_buffer += "iat="
        rst_buffer += str(atoms[3].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[4].pdb_index) + ","
        rst_buffer += str(atoms[5].pdb_index) + "," + end

        rst_buffer += "rstwt="
        rst_buffer += str(weights[1]) + ","
        rst_buffer += str(-weights[1]) + "," + end

        r2 = centre_pair[1]
        r3 = centre_pair[1]

        r1 = centre_pair[1] - harmonic_width
        r4 = centre_pair[1] + harmonic_width

        rk2 = force_constant
        rk3 = force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end + end

        restraints.append(
            {
                "centre_pair": centre_pair,
                "r1": r1,
                "r2": r2,
                "r3": r3,
                "r4": r4,
                "rk2": rk2,
                "rk3": rk3,
            }
        )

        if subdir is not None:
            out_rst = open(subdir + "/window_" + str(i + 1).zfill(2) + ".RST", "w")
            out_rst.write(rst_buffer)
            out_rst.close()
            mout.out(
                "File written to "
                + mcol.file
                + subdir
                + "/window_"
                + str(i + 1).zfill(2)
                + ".RST"
            )

    if adiab_windows:

        for i, centre_pair in enumerate(centres_2d):

            end = "\n"

            rst_buffer = "# adiab_" + str(i + 1) + ".RST" + end

            # First reaction coord

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(atoms[0].pdb_index) + ","
            rst_buffer += str(atoms[1].pdb_index) + ","
            rst_buffer += str(atoms[1].pdb_index) + ","
            rst_buffer += str(atoms[2].pdb_index) + "," + end

            rst_buffer += "rstwt="
            rst_buffer += str(weights[0]) + ","
            rst_buffer += str(-weights[0]) + "," + end

            r2 = centre_pair[0]
            r3 = centre_pair[0]

            r1 = centre_pair[0] - harmonic_width * 10
            r4 = centre_pair[0] + harmonic_width * 10

            rk2 = force_constant * 50
            rk3 = force_constant * 50

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end

            # Second reaction coord

            rst_buffer += "&rst" + end

            rst_buffer += "iat="
            rst_buffer += str(atoms[3].pdb_index) + ","
            rst_buffer += str(atoms[4].pdb_index) + ","
            rst_buffer += str(atoms[4].pdb_index) + ","
            rst_buffer += str(atoms[5].pdb_index) + "," + end

            rst_buffer += "rstwt="
            rst_buffer += str(weights[1]) + ","
            rst_buffer += str(-weights[1]) + "," + end

            r2 = centre_pair[1]
            r3 = centre_pair[1]

            r1 = centre_pair[1] - harmonic_width * 10
            r4 = centre_pair[1] + harmonic_width * 10

            rk2 = force_constant * 50
            rk3 = force_constant * 50

            rst_buffer += "r1=" + str(r1) + "," + end
            rst_buffer += "r2=" + str(r2) + "," + end
            rst_buffer += "r3=" + str(r3) + "," + end
            rst_buffer += "r4=" + str(r4) + "," + end
            rst_buffer += "rk2=" + str(rk2) + "," + end
            rst_buffer += "rk3=" + str(rk3) + "," + end

            rst_buffer += "/" + end + end

            if subdir is not None:
                out_rst = open(subdir + "/adiab_" + str(i + 1).zfill(2) + ".RST", "w")
                out_rst.write(rst_buffer)
                out_rst.close()
                mout.out(
                    "File written to "
                    + mcol.file
                    + subdir
                    + "/adiab_"
                    + str(i + 1).zfill(2)
                    + ".RST"
                )


# import mplot

# xdata = []
# big_ydata = []

# for i in range(samples):

# 	# print(i,coord_range[0]+i*(coord_range[1]-coord_range[0])/(samples-1))
# 	x = coord_range[0]+(i*1.4/(samples-1)-0.2)*(coord_range[1]-coord_range[0])
# 	xdata.append(x)

# for restraint in restraints:

# 	ydata = []

# 	r1 = restraint['r1']
# 	r2 = restraint['r2']
# 	r3 = restraint['r3']
# 	r4 = restraint['r4']
# 	rk2 = restraint['rk2']
# 	rk3 = restraint['rk3']

# 	for x in xdata:

# 		y = restraint_potential(x,r1,r2,r3,r4,rk2,rk3)

# 		ydata.append(y)

# 	big_ydata.append(ydata)

# # if graph:
# if subdir is not None:
# 	graphfile=subdir+"/allwindows.png"
# else:
# 	graphfile=None

# if subdir is not None and graph is not None:
# 	mplot.graph2D(xdata,big_ydata,show=graph,ymax=120,ymin=-5,filename=graphfile)

# if subdir is not None:

# 	pot_buffer = "# restraint potentials"+end

# 	pot_buffer += str(x) + " "

# 	for restraint in restraints:
# 		r1 = restraint['r1']
# 		r2 = restraint['r2']
# 		r3 = restraint['r3']
# 		r4 = restraint['r4']
# 		rk2 = restraint['rk2']
# 		rk3 = restraint['rk3']
# 		y = restraint_potential(x,r1,r2,r3,r4,rk2,rk3)
# 		pot_buffer += str(y) + " "

# 	out_dat = open(subdir+"/allwindows.dat","w")
# 	out_dat.write(pot_buffer)
# 	out_dat.close()


def restraint_potential(x, r1, r2, r3, r4, rk2, rk3):
    if x < r1:
        return lin1(x, r1, r2, rk2)

    if x >= r1 and x <= r2:
        return par2(x, r2, rk2)

    if x >= r3 and x <= r4:
        return par3(x, r3, rk3)

    if x > r4:
        return lin4(x, r3, r4, rk3)

    return 0.0


def lin1(x, r1, r2, rk2):
    return 2 * rk2 * (r1 - r2) * (x - r1) + par2(r1, r2, rk2)


def lin4(x, r3, r4, rk3):
    return 2 * rk3 * (r4 - r3) * (x - r4) + par3(r4, r3, rk3)


def par2(x, r2, rk2):
    return pow(rk2 * (x - r2), 2)


def par3(x, r3, rk3):
    return pow(rk3 * (x - r3), 2)


def prep4amber(system):
    import mcol
    import mout

    # Fix histidine residue names
    if "HSD" in str(system.residues):
        i = system.rename_residues("HSD", "HID", verbosity=0)
        mout.out(
            "Fixed names of "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "HSD"
        )

    ###### Fix names common to many animo acids

    # HN -> H
    aa_hnfix_residues = [
        "ALA",
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "GLY",
        "HID",
        "ILE",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "SER",
        "THR",
        "TRP",
        "TYR",
        "VAL",
    ]

    i = 0
    for resname in aa_hnfix_residues:
        i += system.rename_atoms("HN", "H", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HN->H"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    # HBN->HBN+1
    aa_hbfix_residues = [
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "HID",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "PRO",
        "SER",
        "TRP",
        "TYR",
    ]

    i = 0
    for resname in aa_hbfix_residues:
        system.rename_atoms("HB2", "HB3", res_filter=resname, verbosity=0)
        i += system.rename_atoms("HB1", "HB2", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HBn->HBn+1"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    # HGN->HGN+1
    i = 0
    for resname in ["ALA", "ARG", "GLN", "GLU", "LYS", "MET", "PRO"]:
        system.rename_atoms("HG2", "HG3", res_filter=resname, verbosity=0)
        i += system.rename_atoms("HG1", "HG2", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HGn->HGn+1"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    # HDN->HDN+1
    i = 0
    for resname in ["ARG", "LYS", "PRO"]:
        system.rename_atoms("HD2", "HD3", res_filter=resname, verbosity=0)
        i += system.rename_atoms("HD1", "HD2", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HDn->HDn+1"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    ###### Fix atom names in specific amino acids

    # Fix LEU atom names
    if "LEU" in str(system.residues):
        system.rename_atoms("HT1", "H1", res_filter="LEU", verbosity=0)
        system.rename_atoms("HT2", "H2", res_filter="LEU", verbosity=0)
        i = system.rename_atoms("HT3", "H3", res_filter="LEU", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "LEU"
        )

    # Fix SER atom names
    if "SER" in str(system.residues):
        i = system.rename_atoms("HG1", "HG", res_filter="SER", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "SER"
        )

    # Fix LYS atom names
    if "LYS" in str(system.residues):
        system.rename_atoms("HE2", "HE3", res_filter="LYS", verbosity=0)
        i = system.rename_atoms("HE1", "HE2", res_filter="LYS", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "LYS"
        )

    # Fix GLY atom names
    if "GLY" in str(system.residues):
        system.rename_atoms("HA2", "HA3", res_filter="GLY", verbosity=0)
        i = system.rename_atoms("HA1", "HA2", res_filter="GLY", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "GLY"
        )

    # Fix ILE atom names
    if "ILE" in str(system.residues):
        system.rename_atoms("HG12", "HG13", res_filter="ILE", verbosity=0)
        system.rename_atoms("HG11", "HG12", res_filter="ILE", verbosity=0)
        # system.rename_atoms("2HG1","HG13",res_filter="ILE",verbosity=1)
        # system.rename_atoms("1HG1","HG12",res_filter="ILE",verbosity=1)
        system.rename_atoms("CD", "CD1", res_filter="ILE", verbosity=0)
        system.rename_atoms("HD1", "HD11", res_filter="ILE", verbosity=0)
        system.rename_atoms("HD2", "HD12", res_filter="ILE", verbosity=0)
        i = system.rename_atoms("HD3", "HD13", res_filter="ILE", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "ILE"
        )

    # Fix CYS atom names
    if "CYS" in str(system.residues):
        i = system.rename_atoms("HG1", "HG", res_filter="CYS", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "CYS"
        )

    # Fix ARG atom names
    if "ARG" in str(system.residues):
        system.rename_atoms("OT1", "O", res_filter="ARG", verbosity=0)
        i = system.rename_atoms("OT2", "OXT", res_filter="ARG", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "ARG"
        )

    # Fix thymine atom names
    if "DT" in str(system.residues):
        system.rename_atoms("H51", "H71", res_filter="DT", verbosity=0)
        system.rename_atoms("H52", "H72", res_filter="DT", verbosity=0)
        i = system.rename_atoms("H53", "H73", res_filter="DT", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "DT"
        )

    # HCAP
    i = 0
    for resname in ["DC3", "DA3", "DT3", "DG3"]:
        i += system.rename_atoms("HCAP", "HO3'", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HCAP->HO3'"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of 3TER"
    )

    # Fix ATP atom names
    if "ATP" in str(system.residues):
        system.rename_atoms("O5'", "O5*", res_filter="ATP", verbosity=0)
        system.rename_atoms("C5'", "C5*", res_filter="ATP", verbosity=0)
        system.rename_atoms("C4'", "C4*", res_filter="ATP", verbosity=0)
        system.rename_atoms("O4'", "O4*", res_filter="ATP", verbosity=0)
        system.rename_atoms("C1'", "C1*", res_filter="ATP", verbosity=0)
        system.rename_atoms("C3'", "C3*", res_filter="ATP", verbosity=0)
        system.rename_atoms("O3'", "O3*", res_filter="ATP", verbosity=0)
        system.rename_atoms("C2'", "C2*", res_filter="ATP", verbosity=0)
        system.rename_atoms("O2'", "O2*", res_filter="ATP", verbosity=0)
        system.rename_atoms("H4'", "H40", res_filter="ATP", verbosity=0)
        system.rename_atoms("H1'", "H10", res_filter="ATP", verbosity=0)
        system.rename_atoms("H5'", "H50", res_filter="ATP", verbosity=0)
        system.rename_atoms("H5''", "H51", res_filter="ATP", verbosity=0)
        system.rename_atoms("H8", "H80", res_filter="ATP", verbosity=0)
        system.rename_atoms("H61", "H60", res_filter="ATP", verbosity=0)
        system.rename_atoms("H62", "H61", res_filter="ATP", verbosity=0)
        system.rename_atoms("H2''", "H20", res_filter="ATP", verbosity=0)
        i = system.rename_atoms("H3T", "H30", res_filter="ATP", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "ATP"
        )


def amber2charmm(system):
    import mcol
    import mout

    # Fix histidine residue names
    if "HSD" in str(system.residues):
        i = system.rename_residues("HSD", "HID", verbosity=0)
        mout.out(
            "Fixed names of "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "HSD"
        )

    ###### Fix names common to many animo acids

    # H -> HN
    aa_hnfix_residues = [
        "ALA",
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "GLY",
        "HID",
        "ILE",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "SER",
        "THR",
        "TRP",
        "TYR",
        "VAL",
    ]

    i = 0
    for resname in aa_hnfix_residues:
        i += system.rename_atoms("H", "HN", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "H->HN"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    # HBN->HBN-1
    aa_hbfix_residues = [
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "HID",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "PRO",
        "SER",
        "TRP",
        "TYR",
    ]

    i = 0
    for resname in aa_hbfix_residues:
        i += system.rename_atoms("HB2", "HB1", res_filter=resname, verbosity=0)
        system.rename_atoms("HB3", "HB2", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HBn->HBn-1"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    # HGN->HGN-1
    i = 0
    for resname in ["ALA", "ARG", "GLN", "GLU", "LYS", "MET", "PRO"]:
        i += system.rename_atoms("HG2", "HG1", res_filter=resname, verbosity=0)
        system.rename_atoms("HG3", "HG2", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HGn->HGn-1"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    # HDN->HDN-1
    i = 0
    for resname in ["ARG", "LYS", "PRO"]:
        i += system.rename_atoms("HD2", "HD1", res_filter=resname, verbosity=0)
        system.rename_atoms("HD3", "HD2", res_filter=resname, verbosity=0)
    mout.out(
        mcol.arg
        + "HDn->HDn-1"
        + mcol.clear
        + " rename performed in "
        + mcol.result
        + str(i)
        + mcol.clear
        + " instances of amino acids"
    )

    ###### Fix atom names in specific amino acids

    # Fix LEU atom names
    if "LEU" in str(system.residues):
        system.rename_atoms("H1", "HT1", res_filter="LEU", verbosity=0)
        system.rename_atoms("H2", "HT2", res_filter="LEU", verbosity=0)
        i = system.rename_atoms("H3", "HT3", res_filter="LEU", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "LEU"
        )

    # Fix SER atom names
    if "SER" in str(system.residues):
        i = system.rename_atoms("HG", "HG1", res_filter="SER", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "SER"
        )

    # Fix LYS atom names
    if "LYS" in str(system.residues):
        i = system.rename_atoms("HE2", "HE1", res_filter="LYS", verbosity=0)
        system.rename_atoms("HE3", "HE2", res_filter="LYS", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "LYS"
        )

    # Fix GLY atom names
    if "GLY" in str(system.residues):
        i = system.rename_atoms("HA2", "HA1", res_filter="GLY", verbosity=0)
        system.rename_atoms("HA3", "HA2", res_filter="GLY", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "GLY"
        )

    # Fix ILE atom names
    if "ILE" in str(system.residues):
        system.rename_atoms("HG12", "HG11", res_filter="ILE", verbosity=0)
        system.rename_atoms("HG13", "HG12", res_filter="ILE", verbosity=0)
        # system.rename_atoms("2HG1","HG13",res_filter="ILE",verbosity=1)
        # system.rename_atoms("1HG1","HG12",res_filter="ILE",verbosity=1)
        system.rename_atoms("CD1", "CD", res_filter="ILE", verbosity=0)
        system.rename_atoms("HD11", "HD1", res_filter="ILE", verbosity=0)
        system.rename_atoms("HD12", "HD2", res_filter="ILE", verbosity=0)
        i = system.rename_atoms("HD13", "HD3", res_filter="ILE", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "ILE"
        )

    # Fix CYS atom names
    if "CYS" in str(system.residues):
        i = system.rename_atoms("HG", "HG1", res_filter="CYS", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "CYS"
        )

    # Fix ARG atom names
    if "ARG" in str(system.residues):
        system.rename_atoms("OT1", "O", res_filter="ARG", verbosity=0)
        i = system.rename_atoms("OXT", "OT2", res_filter="ARG", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "ARG"
        )

    # Fix thymine atom names
    if "DT" in str(system.residues):
        system.rename_atoms("H71", "H51", res_filter="DT", verbosity=0)
        system.rename_atoms("H72", "H52", res_filter="DT", verbosity=0)
        i = system.rename_atoms("H73", "H53", res_filter="DT", verbosity=0)
        mout.out(
            "Fixed atom names in "
            + mcol.result
            + str(i)
            + mcol.clear
            + " instances of "
            + mcol.arg
            + "DT"
        )


# # Fix HIS atom names
# if "HIS" in str(system.residues):
# 	i = system.rename_atoms("HD2","HD1",res_filter="HIS",verbosity=0)
# 	mout.out("Fixed atom names in "+mcol.result+str(i)+mcol.clear+" instances of "+mcol.arg+"HIS")

# # Fix ATP atom names
# if "ATP" in str(system.residues):
# 	system.rename_atoms("O5'","O5*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("C5'","C5*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("C4'","C4*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("O4'","O4*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("C1'","C1*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("C3'","C3*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("O3'","O3*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("C2'","C2*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("O2'","O2*",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H4'","H40",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H1'","H10",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H5'","H50",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H5''","H51",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H8","H80",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H61","H60",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H62","H61",res_filter="ATP",verbosity=0)
# 	system.rename_atoms("H2''","H20",res_filter="ATP",verbosity=0)
# 	i = system.rename_atoms("H3T","H30",res_filter="ATP",verbosity=0)
# 	mout.out("Fixed atom names in "+mcol.result+str(i)+mcol.clear+" instances of "+mcol.arg+"ATP")


def parseRST(filename, key, convert_to_float=True):
    import re

    file = open(filename, "r")

    output = []

    for line in file:
        result = re.search(key + "=(.*),", line)
        if result is not None:
            value = result.group(1)
            if convert_to_float:
                value = float(value)
            output.append(value)

        # print(result.group(1))

    return output



import numpy as np


def random_point_spherical(origin, radius):
    while True:

        x = np.random.uniform(-1.0, 1.0)
        y = np.random.uniform(-1.0, 1.0)
        z = np.random.uniform(-1.0, 1.0)

        if np.linalg.norm([x, y, z]) < 1:
            return origin + radius * np.array([x, y, z])



def get_link_atom(site, reference_atom, l=1.0):
    """Get a hydrogen link atom positioned at l Å away from site towards reference_atom"""
    import numpy as np
    from ..atom import Atom

    direction = reference_atom - site
    new_position = site + l * (direction / np.linalg.norm(direction))

    atom = Atom(
        "H",
        position=list(new_position),
        chain=site.chain,
        residue=site.residue,
        res_number=site.res_number,
        res_index=site.res_index,
        occupancy=1.0,
        temp_factor=0.0,
    )
    # atom = Atom("H",position=list(new_position))

    return atom



from .rand import random_point_spherical
import numpy as np


def mc_spherical(
    volume,
    radius: float,
    proximity: float,
    origin: np.ndarray = None,
    samples: int = 4000,
    boolean: bool = False,
    inverse: bool = False,
    points: bool = False,
):
    """
    volume: CompoundVolume object
    radius: radius of testing sphere
    proximity: value to pass to volume.is_within
    samples: number of MC samples
    boolean: return True/False if any valid point in testing sphere
    inverse: points within the volume are invalid
    """

    num_in = 0
    num_out = 0
    points_in = []
    points_out = []

    if origin is None:
        origin = volume.origin

    for i in range(samples):
        point = random_point_spherical(origin, radius)

        if volume.is_within(point, proximity):

            # point is in volume

            if inverse:
                num_out += 1
                points_out.append(point)
            else:
                num_in += 1
                if boolean:
                    return True
                points_in.append(point)
        else:

            # point is not in volume

            if inverse:
                num_in += 1
                if boolean:
                    return True
                points_in.append(point)
            else:
                num_out += 1
                points_out.append(point)

    # fig = volume.plot(radius)

    # import mgo

    # print(f'{point=}')
    # print(f'{points_in=}')
    # print(f'{points_out=}')

    # fig.add_trace(mgo.point_trace(point))

    # fig.show()

    # exit()

    if boolean:
        return False

    sphere_volume = np.pi * np.power(radius, 3)
    fraction = num_out / samples
    valid_volume = fraction * sphere_volume

    if points:
        return fraction, valid_volume, points_in, points_out
    else:
        return fraction, valid_volume



def apply_rototranslation(positions, source_center, target_center, rotation_matrix):
    from numpy import dot

    return dot(positions - source_center, rotation_matrix.T) + target_center


def compute_rototranslation_matrix(source_coords, target_coords):
    """Compute the rototranslation matrix required to align one protein to another"""
    # rotation_matrix_from_points(pself.T, ptarget.T)

    raise NotImplementedError

    from numpy import dot, mean, matmul, array, double
    from ase.build.rotate import rotation_matrix_from_points

    source_center = mean(source_coords, axis=0)
    source_coords -= source_center

    target_center = mean(target_coords, axis=0)
    target_coords -= target_center

    # get the rotation matrix
    rotation = rotation_matrix_from_points(source_coords.T, target_coords.T)

    # matrix equivalent?

    # dot(positions - source_center, rotation_matrix.T) + target_center

    transformation_1 = [
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [*(-source_center), 1],
    ]

    transformation_2 = [
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [*(target_center), 1],
    ]

    transformation_1 = array(transformation_1, dtype=double)
    transformation_2 = array(transformation_2, dtype=double)
    rotation = array(rotation, dtype=double)
    # return matmul(transformation_2.T, transformation_1.T)

    # final = transformation_2 * rotation * transformation_1

    rotation = [
        [*rotation[0], 0],
        [*rotation[1], 0],
        [*rotation[2], 0],
        [0, 0, 0, 1],
    ]

    return transformation_1

    # print(rotation)
    # print(transformation_1)

    step_1 = matmul(rotation.T, transformation_1.T)
    step_2 = matmul(transformation_2.T, step_1)

    return step_2


def apply_transformation(positions, matrix):
    from numpy import matmul, mean

    positions = add_unity_column(positions)
    # print(mean(positions, axis=0))
    positions = matmul(matrix.T, positions.T)
    # print(mean(positions.T, axis=0))
    return positions[:-1, :].T
    # return dot(positions, matrix.T)


def add_unity_column(positions):
    from numpy import c_, ones  # , mean

    # print(positions)
    positions = c_[positions, ones(positions.shape[0])]
    # print(positions)
    # print(positions[:,:-1])
    return positions


# def remove_last_column(positions):



"""Undocumented. Speak to Max"""


def getAtomLabel(trajectory, index, full=False):
    atom_symbols = trajectory[0].get_chemical_symbols()
    atom_tags = trajectory[0].get_tags()

    atom_symbol = atom_symbols[index]
    if atom_tags[index] != 0:
        atom_tag = str(atom_tags[index])
    else:
        atom_tag = ""

    atom_label = atom_symbol + str(atom_tag) + "[" + str(index) + "]"

    if full:
        return atom_label, atom_symbol, atom_tag
    else:
        return atom_label


def getBondLabel(trajectory, index_pair, full=False):
    index1 = index_pair[0]
    index2 = index_pair[1]

    atom_label1, atom_symbol1, atom_tag1 = getAtomLabel(trajectory, index1, full=True)
    atom_label2, atom_symbol2, atom_tag2 = getAtomLabel(trajectory, index2, full=True)

    bond_title = (
        atom_symbol1
        + str(atom_tag1)
        + "-"
        + atom_symbol2
        + str(atom_tag2)
        + " bond ["
        + str(index1)
        + "-"
        + str(index2)
        + "]"
    )

    if full:
        return bond_title, atom_label1, atom_label2
    else:
        return bond_title


def bondLengthStats(
    trajectory,
    index_pair,
    printScript=False,
    verbosity=1,
    timestep=None,
    yUnit="Angstroms",
    fitMin=None,
    fitMax=None,
    returnData=False,
    dataFile=None,
):
    import mcol
    import mout
    import mplot

    index1 = index_pair[0]
    index2 = index_pair[1]

    bond_title = getBondLabel(trajectory, index_pair)

    xdata = []
    ydata = []

    if timestep is None:
        xlab = "MD Steps"
        xUnit = "MD Step"
    else:
        mout.warningOut(f"Scaling xdata by {timestep}x")
        xlab = "Time [fs]"
        xUnit = "femtoseconds"

    if len(trajectory) > 1:

        for n, atoms in enumerate(trajectory):

            if timestep is None:
                xdata.append(n)
            else:
                xdata.append(n * timestep)

            dist = atoms.get_distance(index1, index2)

            ydata.append(dist)

        val, err, fit_func = mplot.fit(
            xdata,
            ydata,
            rank=0,
            verbosity=verbosity - 1,
            fitMin=fitMin,
            fitMax=fitMax,
            title=bond_title,
            yUnit=yUnit,
            xUnit=xUnit,
        )

    else:

        val = trajectory[0].get_distance(index1, index2)
        err = None

    mout.varOut(
        bond_title,
        val,
        error=err,
        unit=yUnit,
        valCol=mcol.result,
        dataFile=dataFile,
        verbosity=verbosity,
    )

    if returnData:
        return val, err, bond_title, xdata, ydata
    else:
        return val, err, bond_title


def bondAngleStats(
    trajectory,
    index_triplet,
    printScript=False,
    verbosity=1,
    timestep=None,
    fitMin=None,
    fitMax=None,
    yUnit="degrees",
    returnData=False,
    dataFile=None,
):
    import mcol
    import mout
    import mplot
    from ase.units import fs

    atom_symbols = trajectory[0].get_chemical_symbols()
    atom_tags = trajectory[0].get_tags()

    index1 = index_triplet[0]
    index2 = index_triplet[1]
    index3 = index_triplet[2]

    atom_symbol1 = atom_symbols[index1]
    atom_symbol2 = atom_symbols[index2]
    atom_symbol3 = atom_symbols[index3]

    if atom_tags[index1] != 0:
        atom_tag1 = str(atom_tags[index1])
    else:
        atom_tag1 = ""
    if atom_tags[index2] != 0:
        atom_tag2 = str(atom_tags[index2])
    else:
        atom_tag2 = ""
    if atom_tags[index3] != 0:
        atom_tag3 = str(atom_tags[index3])
    else:
        atom_tag3 = ""

    atom_label1 = atom_symbol1 + str(atom_tag1) + "[" + str(index1) + "]"
    atom_label2 = atom_symbol2 + str(atom_tag2) + "[" + str(index2) + "]"
    atom_label3 = atom_symbol3 + str(atom_tag3) + "[" + str(index3) + "]"
    labels = [atom_label1, atom_label2, atom_label3]
    bond_title = (
        atom_symbol1
        + str(atom_tag1)
        + "-"
        + atom_symbol2
        + str(atom_tag2)
        + "-"
        + atom_symbol3
        + str(atom_tag3)
        + " bond ["
        + str(index1)
        + "-"
        + str(index2)
        + "-"
        + str(index3)
        + "]"
    )

    xdata = []
    ydata = []

    if timestep is None:
        xlab = "MD Steps"
        xUnit = "MD Step"
    else:
        xlab = "Time [fs]"
        xUnit = "picosecond"

    if len(trajectory) > 1:

        for n, atoms in enumerate(trajectory):
            if timestep is None:
                xdata.append(n)
            else:
                xdata.append(n * timestep / fs)

            ang = atoms.get_angle(index1, index2, index3)

            ydata.append(ang)

        val, err, fit_func = mplot.fit(
            xdata,
            ydata,
            rank=0,
            verbosity=verbosity - 1,
            fitMin=fitMin,
            fitMax=fitMax,
            title=bond_title,
            yUnit=yUnit,
            xUnit=xUnit,
        )

    else:
        val = trajectory[0].get_angle(index1, index2, index3)
        err = None

    mout.varOut(
        bond_title,
        val,
        error=err,
        unit=yUnit,
        valCol=mcol.result,
        dataFile=dataFile,
        verbosity=verbosity,
    )

    if returnData:
        return val, err, bond_title, xdata, ydata
    else:
        return val, err, bond_title


def getCentreOfMass(atoms, verbosity=1):
    positions = getPositions(atoms)

    this_com = [
        sum([pos[0] for pos in positions]) / len(positions),
        sum([pos[1] for pos in positions]) / len(positions),
        sum([pos[2] for pos in positions]) / len(positions),
    ]

    if verbosity > 0:
        mout.varOut("CoM (" + str(atoms) + ")", this_com, unit="Å")

    return this_com


def getRMSD(atoms, reference=None, verbosity=1):
    import mout

    if reference is None:
        CoM = getCentreOfMass(atoms, verbosity=verbosity - 1)
    else:
        ref_positions = getPositions(reference)

    positions = getPositions(atoms)

    num_atoms = len(positions)

    big_sum = 0.0

    for index, pos in enumerate(positions):
        if reference is None:
            d = getDisplacement(pos, CoM)
        else:
            d = getDisplacement(pos, ref_positions[index])

        big_sum += pow(d, 2)

    rmsd = pow(big_sum / num_atoms, 0.5)

    mout.varOut("RMSD (" + str(atoms) + ")", rmsd, unit="Å")

    return rmsd


def getDisplacement(position1, position2):
    x = position1[0] - position2[0]
    y = position1[1] - position2[1]
    z = position1[2] - position2[2]
    d = pow(pow(x, 2) + pow(y, 2) + pow(z, 2), 0.5)
    return d


def getAngle(position1, position2, position3):
    import numpy as np

    a = np.array(position1)
    b = np.array(position2)
    c = np.array(position3)

    ba = a - b
    bc = c - b

    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
    angle = np.arccos(cosine_angle)

    return np.degrees(angle)


def getPositions(atoms):
    from ase import Atoms

    if isinstance(atoms, Atoms):
        # ASE ATOMS
        positions = atoms.get_positions()
    elif isinstance(atoms, list):
        positions = []
        for thing in atoms:
            positions += thing.positions
    else:
        # AMP System/Chain/Residue
        positions = atoms.positions

    return positions



def pdb2traj(input, output, verbosity=1, printScript=False, tagging=True):
    import mout
    import mcol
    from .io import read, write

    if not input.endswith(".pdb"):
        mout.errorOut("Input is not a PDB!")
        return None

    if verbosity > 0:
        mout.out(
            "Converting "
            + mcol.file
            + input
            + mcol.clear
            + " to "
            + mcol.file
            + output
            + mcol.clear
            + " ... ",
            printScript=printScript,
            end="",
        )
        if verbosity > 1:
            mout.out("")

        in_traj = read(input, index=":", verbosity=verbosity - 1, tagging=tagging)

        write(output, in_traj, verbosity=verbosity - 1)

        if verbosity == 1:
            mout.out("Done.")

        return in_traj


def gro2traj(input, output, verbosity=1, printScript=False, tagging=True):
    import mout
    import mcol
    from .io import read, write

    if not input.endswith(".gro"):
        mout.errorOut("Input is not a .gro file!")
        return None

    if verbosity > 0:
        mout.out(
            "Converting "
            + mcol.file
            + input
            + mcol.clear
            + " to "
            + mcol.file
            + output
            + mcol.clear
            + " ... ",
            printScript=printScript,
            end="",
        )
        if verbosity > 1:
            mout.out("")

    if tagging:
        # initialise arrays
        taglist = []

        if verbosity > 1:
            mout.out("parsing GRO for tags ... ", printScript=printScript, end="")

        line_number = 0
        max_line_number = 1000

        # Parse the GRO to get the tags
        with open(input, "r") as input_gro:
            for line in input_gro:
                if line_number == 1:
                    max_line_number = int(line.strip().split()[0]) + 2
                if line_number == max_line_number:
                    break
                if line_number > 1:
                    taglist.append(
                        int(
                            "".join(
                                filter(lambda i: i.isdigit(), line.strip().split()[1])
                            )
                        )
                    )
                line_number = line_number + 1

        if verbosity > 1:
            mout.out("Done.")

    in_traj = read(input, index=":", verbosity=verbosity - 1)

    if tagging:
        # set the tags
        for atoms in in_traj:
            for index, tag in enumerate(taglist):
                atoms[index].tag = tag

    write(output, in_traj, verbosity=verbosity - 1)

    if verbosity == 1:
        mout.out("Done.")

    return in_traj


def xyz2traj(input, output, verbosity=1, printScript=False):
    import mout
    import mcol
    from .io import read, write

    if not input.endswith(".xyz"):
        mout.errorOut("Input is not a .xyz file!")
        return None

    if verbosity > 0:
        mout.out(
            "Converting "
            + mcol.file
            + input
            + mcol.clear
            + " to "
            + mcol.file
            + output
            + mcol.clear
            + " ... ",
            printScript=printScript,
            end="",
        )
        if verbosity > 1:
            mout.out("")

    in_traj = read(input, index=":", verbosity=verbosity - 1)

    write(output, in_traj, verbosity=verbosity - 1)

    if verbosity == 1:
        mout.out("Done.")

    return in_traj


def fileLength(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1



#!/usr/bin/env python3

from mout import debug_log
from mout.testing import TestStatus

import mrich as logger


from molparse.residue import add_res_type, res_type


@debug_log
def check_res_type(res_name):
    test_status = TestStatus()

    logger.print(f"{res_type(res_name)=}")

    return test_status


def main():

    check_res_type("DG")
    check_res_type("TYR")
    check_res_type("CRO")
    check_res_type("CROW")
    check_res_type("COW")
    check_res_type("DOG")

    add_res_type("COW", "PRO")
    add_res_type("CROW", "LIP")
    add_res_type("DOG", "LIG")

    check_res_type("CROW")
    check_res_type("COW")
    check_res_type("DOG")


if __name__ == "__main__":
    main()



class Network(object):
    """Barebones class for visualising a network/graph"""

    def __init__(self, connections):

        import networkx as nx

        self._connections = connections
        self._graph = nx.Graph()

        for u, v in connections:
            self._graph.add_edge(u, v)

    def relax_pos(self, seed=None):
        import networkx as nx

        if seed:
            return nx.spring_layout(self._graph, seed=seed)
        else:
            return nx.spring_layout(self._graph)

    def plot(self, seed=None, show=True):

        import networkx as nx
        import matplotlib.pyplot as plt

        pos = self.relax_pos(seed)

        options = {
            "font_size": 36,
            "node_size": 3000,
            "node_color": "white",
            "edgecolors": "black",
            "linewidths": 5,
            "width": 5,
        }

        # draw network
        nx.draw_networkx(self._graph, pos, **options)

        ax = plt.gca()
        ax.margins(0.2)
        plt.axis("off")

        if show:
            plt.show()

        return ax



isPovLoaded = False


def loadPov(verbosity=1, purge=True, anaconda=False):
    import module  # https://github.com/mwinokan/MPyTools

    # load the correct modules for ASE/PoV-Ray
    if purge:
        module.module("purge")
    module.module("--expert", "load", "Boost/1.63.0-intel-2017a-Python-2.7.13")
    module.module("--expert", "load", "zlib/1.2.8-intel-2016a")
    module.module("--expert", "load", "libpng/1.6.24-intel-2016a")
    module.module("--expert", "load", "libjpeg-turbo/1.5.0-intel-2016a")
    module.module("--expert", "load", "LibTIFF/4.0.6-intel-2016a")
    if anaconda:
        module.module("--expert", "load", "anaconda3/2019.03")

    if verbosity > 0:
        mout.out(
            "PoV-Ray dependencies loaded.",
            printScript=True,
        )

    global isPovLoaded
    isPovLoaded = True


def makePovImage(
    filename,
    image,
    verbosity=1,
    rmPovFiles=True,
    bonds=False,
    bondradius=1.1,
    forceLoad=False,
    printScript=False,
    **style,
):
    from ase import io
    import mcol
    import mout

    if not isPovLoaded or forceLoad:
        loadPov(verbosity=verbosity - 1)

    if verbosity > 0:
        mout.out(
            "processing " + mcol.file + filename + ".pov" + mcol.clear + " ... ",
            printScript=printScript,
            end="",
        )  # user output

    if not style["drawCell"]:
        image.set_cell([0, 0, 0])
    del style["drawCell"]

    if "canvas_height" in style:
        del style["canvas_height"]

    if bonds:
        from ase.io.pov import get_bondpairs, set_high_bondorder_pairs

        bondpairs = get_bondpairs(image, radius=bondradius)
        if len(bondpairs) > 5000:
            mout.warningOut(
                "Too many bondpairs (" + str(len(bondpairs)) + "), not drawing bonds!",
                end=" ",
            )
        else:
            style["bondatoms"] = bondpairs

    io.write(
        filename + ".pov", image, run_povray=True, camera_type="perspective", **style
    )

    if rmPovFiles:
        import os

        os.system("rm " + filename + ".ini")
        os.system("rm " + filename + ".pov")

    if verbosity > 0:
        mout.out("Done.")  # user output


def makePovImages(
    filename,
    subdirectory="pov",
    interval=1,
    verbosity=1,
    rmPovFiles=True,
    bonds=False,
    bondradius=1.1,
    filenamePadding=4,
    printScript=False,
    forceLoad=False,
    index=":",
    **style,
):
    from ase import io
    import mcol
    import mout

    if not isPovLoaded or forceLoad:
        loadPov(verbosity=verbosity - 1)

    import os
    import math

    os.system("mkdir -p " + subdirectory)
    # os.system("rm -v "+subdirectory+"/* ")
    os.system("rm " + subdirectory + "/* 2> /dev/null")

    if index != ":":
        image = io.read(filename, index=index)
        makePovImage(
            subdirectory + "/" + str(index).zfill(filenamePadding),
            image,
            verbosity=verbosity - 1,
            bonds=bonds,
            bondradius=bondradius,
            printScript=printScript,
            rmPovFiles=False,
            forceLoad=forceLoad,
            **style,
        )
    else:
        traj = io.read(filename, index=index)

        global num_traj_images
        global num_frames
        num_traj_images = len(traj)
        num_frames = math.ceil(num_traj_images / interval)
        if verbosity > 0:
            mout.varOut("Trajectory", num_traj_images, unit="images")
            mout.varOut("Animation", num_frames, unit="frames")

        for n, image in enumerate(traj):
            if n % interval != 0 and n != 100:
                continue

            if verbosity == 1:
                mout.progress(
                    n + 1,
                    num_traj_images,
                    prepend="Creating images",
                    printScript=printScript,
                )

            makePovImage(
                subdirectory + "/" + str(n).zfill(filenamePadding),
                image,
                verbosity=verbosity - 1,
                bonds=bonds,
                bondradius=bondradius,
                printScript=printScript,
                rmPovFiles=False,
                forceLoad=forceLoad,
                **style,
            )

    if rmPovFiles:
        os.system("rm " + subdirectory + "/*.ini")
        os.system("rm " + subdirectory + "/*.pov")


# Using ImageMagick (artefacting!):
def makePovAnimationIM(filename, subdirectory="pov", interval=1, verbosity=1, **style):
    import mcol
    import mout

    if not isPovLoaded or forceLoad:
        loadPov(verbosity=verbosity - 1)

    import os
    import module  # https://github.com/mwinokan/MPyTools

    makePovImages(
        filename,
        subdirectory=subdirectory,
        interval=interval,
        verbosity=verbosity - 1,
        **style,
    )

    module.module("--expert", "load", "ImageMagick/7.0.3-1-intel-2016a")
    if verbosity > 0:
        mout.out("ImageMagick loaded.", printScript=True)

    if verbosity > 0:
        mout.out(
            "creating " + mcol.file + "animation.gif" + mcol.clear + " ... ",
            printScript=True,
            end="",
        )  # user output
    os.system(
        "convert -delay 10 "
        + subdirectory
        + "/*.png -fill white -opaque none -loop 1 "
        + subdirectory
        + ".gif"
    )
    if verbosity > 0:
        mout.out("Done.")  # user output


# Using imageio
# https://stackoverflow.com/questions/753190/programmatically-generate-video-or-animated-gif-in-python
def makePovAnimation(
    filename,
    subdirectory="pov",
    interval=1,
    gifstyle=None,
    verbosity=1,
    printScript=False,
    forceLoad=False,
    useExisting=False,
    dryRun=False,
    **plotstyle,
):
    import mcol
    import mout

    if not isPovLoaded or forceLoad:
        loadPov(verbosity=verbosity - 1)

    from . import styles

    if gifstyle is None:
        gifstyle = styles.gif_standard

    if plotstyle == {}:
        plotstyle = styles.standard
    # print(plotstyle)

    global num_traj_images
    global num_frames

    import os
    import imageio

    cropping = False
    shifting = False

    # Set canvas sizes:
    canv_w = plotstyle["canvas_width"]
    if "canvas_height" in plotstyle:
        canv_h = plotstyle["canvas_height"]
        del plotstyle["canvas_height"]
    else:
        mout.errorOut("No canvas_height specified in plotstyle!", fatal=True)

    # Check if cropping:
    if "crop_w" in plotstyle and "crop_w" in plotstyle:
        cropping = True
        crop_w = plotstyle["crop_w"]
        crop_h = plotstyle["crop_h"]
    if "crop_w" in plotstyle:
        del plotstyle["crop_w"]
    if "crop_h" in plotstyle:
        del plotstyle["crop_h"]

    # Check if crop offset:
    if "crop_x" in plotstyle:
        shifting = True
        crop_x = plotstyle["crop_x"]
        crop_y = plotstyle["crop_y"]
    if "crop_x" in plotstyle:
        del plotstyle["crop_x"]
    if "crop_y" in plotstyle:
        del plotstyle["crop_y"]

    mout.varOut("cropping", cropping)
    if cropping:
        mout.varOut("crop_w", crop_w)
        mout.varOut("crop_h", crop_h)
    mout.varOut("shifting", shifting)

    # Generate the PNG's
    if not useExisting:
        if verbosity > 0:
            mout.out(
                "generating "
                + mcol.file
                + subdirectory
                + "/*.png"
                + mcol.clear
                + " ... ",
                printScript=printScript,
                end="",
            )  # user output
        if verbosity > 1:
            mout.out(" ")

        if not dryRun:
            # Generate all the images
            makePovImages(
                filename,
                subdirectory=subdirectory,
                interval=interval,
                printScript=printScript,
                verbosity=verbosity - 1,
                **plotstyle,
            )
        else:
            # Generate just the first image
            makePovImages(
                filename,
                subdirectory=subdirectory,
                interval=interval,
                printScript=printScript,
                verbosity=verbosity - 1,
                index=0,
                **plotstyle,
            )

        if verbosity == 1:
            mout.out("Done.")

    # Load ImageMagick
    # if cropping or backwhite:
    import module  # https://github.com/mwinokan/MPyTools

    ret = module.module("--expert", "load", "ImageMagick/7.0.3-1-intel-2016a")
    if ret == 0 and verbosity > 0:
        mout.out("ImageMagick loaded.", printScript=printScript)

    # Combine the images
    if verbosity > 0:
        mout.out(
            "loading " + mcol.file + subdirectory + "/*.png" + mcol.clear + " ... ",
            printScript=printScript,
            end="",
        )  # user output

    images = []

    # loop over all files in the subdirectory:
    for file in sorted(os.listdir(subdirectory)):

        # get the relative path to the file:
        filename = subdirectory + "/" + file

        # check if the file is a PNG:
        if file.endswith(".png"):

            tempname = subdirectory + "/" + "temp.png"

            # run different IM commands depending on cropping and shifting:
            if not cropping and not shifting:
                # os.system("convert "+filename+
                #           " -background white -extent "+
                #           str(canv_w)+"x"+
                #           str(canv_h)+" "+
                #           filename)
                os.system("convert " + filename + " -flatten " + tempname)
            elif cropping and not shifting:
                os.system(
                    "convert "
                    + filename
                    + " -crop "
                    + str(crop_w)
                    + "x"
                    + str(crop_h)
                    + " -background white -extent "
                    + str(crop_w)
                    + "x"
                    + str(crop_h)
                    + " "
                    + tempname
                )
                print(
                    "convert "
                    + filename
                    + " -crop "
                    + str(crop_w)
                    + "x"
                    + str(crop_h)
                    + " -background white -extent "
                    + str(crop_w)
                    + "x"
                    + str(crop_h)
                    + " "
                    + tempname
                )
                os.system("ls NEB")
            elif shifting and not cropping:
                os.system(
                    "convert "
                    + filename
                    + " -crop +"
                    + str(crop_x)
                    + "+"
                    + str(crop_y)
                    + " -background white -extent "
                    + str(canv_w)
                    + "x"
                    + str(canv_h)
                    + " "
                    + tempname
                )
            else:
                os.system(
                    "convert "
                    + filename
                    + " -crop "
                    + str(crop_w)
                    + "x"
                    + str(crop_h)
                    + "+"
                    + str(crop_x)
                    + "+"
                    + str(crop_y)
                    + " -background white -extent "
                    + str(crop_w)
                    + "x"
                    + str(crop_h)
                    + " "
                    + tempname
                )

            os.system("mv " + tempname + " " + filename)

            # Read in the image and append to the image array
            image = imageio.imread(filename)
            images.append(image)

            if verbosity > 0 and not dryRun:
                mout.progress(
                    len(images),
                    num_frames,
                    prepend="Cropping & loading images",
                    printScript=printScript,
                )

    if (verbosity > 0) and not useExisting:
        mout.out("Done.")  # user output

    if verbosity > 0:
        mout.out(
            "creating " + mcol.file + subdirectory + ".gif" + mcol.clear + " ... ",
            printScript=printScript,
            end="",
        )  # user output

    # Generate the animated GIF:
    imageio.mimsave(subdirectory + ".gif", images, **gifstyle)

    if verbosity > 0:
        mout.out("Done.")  # user output


def crop(filename, width=500, height=500, xshift=0, yshift=0, verbosity=1):
    import mcol
    import mout
    import os

    import module  # https://github.com/mwinokan/MPyTools

    module.module("--expert", "load", "ImageMagick/7.0.3-1-intel-2016a")
    global isPovLoaded
    isPovLoaded = False
    if verbosity > 0:
        mout.out("ImageMagick loaded.", printScript=printScript)

    os.system(
        "convert "
        + filename
        + " -crop "
        + str(width)
        + "x"
        + str(height)
        + "+"
        + str(xshift)
        + "+"
        + str(yshift)
        + " "
        + filename
    )



import mrich as logger


def protonate(
    sys: "System",
    minimise: bool = False,
    pH: float = 7.8,
    remove_residues: list[str] | None = ["DMS", "TRS", "LIG", "CL"],
    trim_terminal_residues: int = 0,
    return_file: bool = False,
) -> "System | str":
    """Protonate a system using pdbfixer and openmm

    :param sys: Input :class:`.System`
    :param minimise: Perform an energy minimisation?
    :param pH: System pH
    :param remove_residues: list of residue names to remove
    :param trim_terminal_residues: number of residues to trim from each end of all protein chains
    :returns: Output :class:`.System`
    """

    from .amber import prep4amber
    from .io import parsePDB
    from openmm.app import PME, ForceField, Modeller, PDBFile, Simulation
    from tempfile import NamedTemporaryFile
    from pdbfixer import PDBFixer

    orig_sys = sys.copy()

    if remove_residues:
        orig_sys.remove_residues(names=remove_residues, no_summary=True)

    if trim_terminal_residues:
        for chain in orig_sys.chains:
            if chain.type != "PRO":
                continue
            residues = [r for r in chain.residues[:trim_terminal_residues]] + [
                r for r in chain.residues[-trim_terminal_residues:]
            ]
            indices = [r.index for r in residues]
            chain.remove_residues(indices=indices)

    # prepare IO files
    pdb_orig = NamedTemporaryFile(mode="w+t", suffix=".pdb")
    pdb_prot = NamedTemporaryFile(mode="w+t", suffix=".pdb")

    # write the original
    orig_sys.write(pdb_orig.name, verbosity=0)

    # fix the PDB termini and hydrogens
    fixer = PDBFixer(pdb_orig.name)
    fixer.findMissingResidues()
    if fixer.missingResidues:
        logger.warning(f"{fixer.missingResidues=}")
    fixer.findNonstandardResidues()
    if fixer.nonstandardResidues:
        logger.warning(f"{fixer.nonstandardResidues=}")
    fixer.findMissingAtoms()
    if fixer.missingAtoms:
        logger.warning(f"{fixer.missingAtoms=}")
    if fixer.missingTerminals:
        logger.print(f"{fixer.missingTerminals=}")
    fixer.addMissingAtoms()

    fixer.addMissingHydrogens(pH)
    PDBFile.writeFile(fixer.topology, fixer.positions, pdb_prot)
    # PDBFile.writeFile(fixer.topology, fixer.positions, "pdb_prot.pdb")

    if minimise:

        pdb_mini = NamedTemporaryFile(mode="w+t", suffix=".pdb")

        # openmm objects
        pdb = PDBFile(pdb_prot.name)
        forcefield = ForceField("amber99sb.xml", "tip3p.xml")
        modeller = Modeller(pdb.topology, pdb.positions)

        # prepare the simulation
        system = forcefield.createSystem(modeller.topology, nonbondedMethod=PME)
        integrator = VerletIntegrator(0.001 * picoseconds)
        simulation = Simulation(modeller.topology, system, integrator)
        simulation.context.setPositions(modeller.positions)

        # minimise proton positions
        simulation.minimizeEnergy(maxIterations=100)

        # update positions
        positions = simulation.context.getState(getPositions=True).getPositions()

        # write files
        PDBFile.writeFile(simulation.topology, positions, pdb_mini)

        # read in mp.System
        sys = parsePDB(pdb_mini.name, verbosity=0)
        pdb_mini.close()

    else:
        # read in mp.System
        sys = parsePDB(pdb_prot.name, verbosity=0)
        pdb_prot.close()

    # fix residue numbering and chain naming
    for orig_chain, new_chain in zip(orig_sys.chains, sys.chains):

        if orig_chain.sequence != new_chain.sequence:

            assert len(orig_chain.sequence) > len(
                new_chain.sequence
            ), f"Sequences don't match: {orig_chain.sequence} {new_chain.sequence}"

            offset = orig_chain.sequence.find(new_chain.sequence)
            assert offset >= 0

        else:
            offset = 0

        new_chain.name = orig_chain.name

        for i, new_residue in enumerate(new_chain.residues):
            orig_residue = orig_chain.residues[i + offset]
            assert orig_residue.name == new_residue.name
            new_residue.number = orig_residue.number

    # close files
    pdb_orig.close()

    if return_file:
        pdb_out = NamedTemporaryFile(mode="w+t", suffix=".pdb")
        sys.write(pdb_out.name, verbosity=0)
        return sys, pdb_out

    return sys



import os
from rdkit import Chem, RDConfig
from rdkit.Chem import AllChem
import numpy as np
from .mol import mol_to_pdb_block
from ..group import AtomGroup

FDEF = AllChem.BuildFeatureFactory(
    os.path.join(RDConfig.RDDataDir, "BaseFeatures.fdef")
)
FEATURE_FAMILIES = list(FDEF.GetFeatureFamilies()) + ["Sulfur"]

COMPLEMENTARY_FEATURES = {
    "Donor": ["Acceptor"],  # hydrogen bond
    "Acceptor": ["Donor"],  # hydrogen bond
    "NegIonizable": ["PosIonizable"],  # electrostatic
    "PosIonizable": ["NegIonizable", "Aromatic"],  # electrostatic, pi-cation
    "Aromatic": ["Aromatic", "PosIonizable"],  # pi-stacking, pi-cation
    "Hydrophobe": ["Hydrophobe", "LumpedHydrophobe"],  # hydrophobic
    "LumpedHydrophobe": ["Hydrophobe", "LumpedHydrophobe"],  # hydrophobic
    "Sulfur": ["Sulfur"],  # sulfur-sulfur
}

INTERACTION_TYPES = {
    ("Hydrophobe", "Hydrophobe"): "Hydrophobic",
    ("LumpedHydrophobe", "Hydrophobe"): "Hydrophobic",
    ("Hydrophobe", "LumpedHydrophobe"): "Hydrophobic",
    ("LumpedHydrophobe", "LumpedHydrophobe"): "Hydrophobic",
    ("Donor", "Acceptor"): "Hydrogen Bond",
    ("Acceptor", "Donor"): "Hydrogen Bond",
    ("NegIonizable", "PosIonizable"): "Electrostatic",
    ("PosIonizable", "NegIonizable"): "Electrostatic",
    ("Aromatic", "Aromatic"): "π-stacking",
    ("Aromatic", "PosIonizable"): "π-cation",
    ("PosIonizable", "Aromatic"): "π-cation",
    ("Sulfur", "Sulfur"): "Sulfur-Sulfur",
}


def features_from_mol(mol, protonate=True, group=None):
    raw_features = raw_features_from_mol(mol, protonate=protonate)

    if not group:
        group = AtomGroup.from_pdb_block(mol_to_pdb_block(mol))

    feature_list = []
    for feat in raw_features:

        indices = feat.GetAtomIds()
        family = feat.GetFamily()

        atoms = [group.atoms[i] for i in indices]

        # position from indices
        if len(indices) == 1:
            position = atoms[0].np_pos
        else:
            position = np.mean([a.np_pos for a in atoms], axis=0)

        # f_dict = dict(family=family,position=position,indices=indices,x=position[0],y=position[1],z=position[2])

        feat_obj = Feature(
            family=family,
            atoms=atoms,
            position=position,
            res_name=None,
            res_number=None,
            res_chain=None,
        )

        feature_list.append(feat_obj)

    ### Add Sulfur

    sulfurs = [atom for atom in group.atoms if atom.symbol == "S"]

    for atom in sulfurs:

        feat_obj = Feature(
            family="Sulfur",
            atoms=[atom],
            position=atom.np_pos,
            res_name=None,
            res_number=None,
            res_chain=None,
        )

        feature_list.append(feat_obj)

    return feature_list


def raw_features_from_mol(mol, protonate=True):
    # protonate
    if protonate:
        m_pdb_prot = Chem.AddHs(mol)
    else:
        m_pdb_prot = mol

    # solve for a pose
    ps = AllChem.ETKDGv3()
    AllChem.EmbedMolecule(m_pdb_prot, ps)

    # feature factory
    fdef = AllChem.BuildFeatureFactory(
        os.path.join(RDConfig.RDDataDir, "BaseFeatures.fdef")
    )

    # get the features
    raw_features = fdef.GetFeaturesForMol(m_pdb_prot)

    return raw_features


def features_from_group(group, protonate=True):
    m_pdb = group.rdkit_mol
    raw_features = raw_features_from_mol(m_pdb, protonate=protonate)

    feature_list = []

    for feat in raw_features:

        indices = feat.GetAtomIds()
        family = feat.GetFamily()

        # position from indices
        if len(indices) == 1:
            atoms = [group.atoms[indices[0]]]
            position = atoms[0].np_pos
        else:
            atoms = [group.atoms[i] for i in indices]
            position = np.mean([group.atoms[i].np_pos for i in indices], axis=0)

        feat_obj = Feature(
            family=family,
            atoms=atoms,
            position=position,
            res_name=None,
            res_number=None,
            res_chain=None,
        )

        feature_list.append(feat_obj)

    return feature_list


import numpy as np


class Feature(object):
    """Object representing an interaction feature (usually on a protein)"""

    def __init__(
        self,
        family: list,
        atoms: list,
        position: np.ndarray,
        res_name: str,
        res_number: int,
        res_chain: str,
    ):

        self.family = family
        self.atoms = atoms
        self.position = position
        self.res_name = res_name
        self.res_number = res_number
        self.res_chain = res_chain

        self.atom_numbers = [a.number for a in self.atoms]

    @property
    def x(self):
        """x coordinate"""
        return self.position[0]

    @property
    def y(self):
        """y coordinate"""
        return self.position[1]

    @property
    def z(self):
        """z coordinate"""
        return self.position[2]

    @property
    def dict(self):
        """Dictionary containing family and coordinates"""
        return dict(
            family=self.family,
            x=self.x,
            y=self.y,
            z=self.z,
        )

    def __repr__(self):
        return f"{self.family} @ {self.x:.2f} {self.y:.2f} {self.z:.2f}"

    @property
    def name_number_chain_str(self):
        """String containing the residue name, residue number, and chain"""
        return f"{self.res_name} {self.res_number} {self.res_chain}"

    @property
    def family_name_number_chain_str(self):
        """String containing the feature family, residue name, residue number, and chain"""
        return f"{self.family} {self.res_name} {self.res_number} {self.res_chain}"

    @property
    def family_name_number_chain_atoms_str(self):
        """String containing the feature family, residue name, residue number, chain, and atoms"""
        return f"{self.family} {self.res_name} {self.res_number} {self.res_chain} {self.atoms}"

    def __sub__(self, other):
        if isinstance(other, np.ndarray):
            return self.position - other
        else:
            return self.position - other.position



"""

========
MolParse
========

A python package for parsing, modifying, and analysis of molecular structure files.

It is recommended to import as the acronym:

        import molparse as mp

Key features
============

Parsing coordinate files into System objects
--------------------------------------------

::

        system = mp.parsePDB(filestr)
        system = mp.parseGRO(filestr)

* Viewing the molecular hierarchy of a system:

        mp.tree(system)

* Parsing of delimited text files:

        see help(mp.signal.parseDat)

* Writing System or ase.Atoms objects or lists thereof:

        mp.write(filestr,object)

* An object oriented approach to molecular system hierarchies:

        mp.System
            |__ mp.Chain
                    |__ mp.Residue
                            |__ mp.Atom

        See also:

        help(mp.System), help(mp.Chain), etc.

* Wrappers for common ASE methods (with mp class support):

        mp.read --> ase.io.read
        mp.write --> ase.io.write
        mp.view --> ase.visualize.view

* Functions to render images and animations of ASE Atoms:

        with matplotlib:
                mp.makeImage,
                mp.makeImages,
                mp.makeAnimation

        with PoVRay:
                mp.makePovImage,
                mp.makePovImages,
                mp.makePovAnimation

* Methods relating to DNA:

        see help(mp.dna)

* Finding the stationary points of a double well potential

        see help(mp.tunnel.find_barrier_stationary_points)

* Graphing of common molecular properties:

        see help(mp.graphing)

* Analysis of common molecular properties:

        see help(mp.analysis)

"""

__version__ = "0.0.31"

# from .version import import_checks
# import_checks()

# Custom IO
from .io import read
from .io import write
from .io import parse
from .io import parsePDB
from .io import writePDB
from .io import parseGRO
from .io import parseXYZ
from .io import modifyPDB

from .signal import parseDat

# CLI-Viewer
# from .tree import tree

# Manipulation
from .manipulate import interpolate
from .manipulate import auto_rotate

# Custom Classes
from .group import AtomGroup
from .system import System
from .chain import Chain
from .residue import Residue
from .atom import Atom
from .restraint import Restraint
from .network import Network

# ase.visualize wrappers
from .gui import view
from .go import plot3d

# ase standard rendering
from .plot import makeImage
from .plot import makeImages
from .plot import makeAnimation

# PoV-Ray rendering
from .povplot import loadPov
from .povplot import makePovImage
from .povplot import makePovImages
from .povplot import makePovAnimation
from .povplot import crop

# Analysis
from .analysis import bondLengthStats
from .analysis import bondAngleStats
from .analysis import getCentreOfMass
from .analysis import getRMSD
from .analysis import getAngle
from .analysis import getDisplacement
from .analysis import getBondLabel
from .analysis import getAtomLabel

# Structure Comparison
# from .compare import compareSystems
from .compare import euclid_dist

# Graphing
from .graphing import showFigs
from .graphing import graphEnergy
from .graphing import graphForces
from .graphing import graphDisplacement
from .graphing import graphBondLength
from .graphing import graphBondAngle
from .graphing import graphBondVibSpec

# Console IO
from .console_io import printEnergy  # needs documentation

# XVG Parsing
from .xvg import parseXVG

# NDX Parsing
from .ndx import parseNDX

# Amber specific
from .amber import prep4amber
from .amber import amber2charmm
from .amber import parseRST
from .amber import umbrella_helper_2dist
from .amber import umb_rst_2prot
from .amber import umb_rst_2prot_new
from .amber import umb_rst_2prot_1rc
from .amber import umb_rst_2prot_asy
from .amber import umbrella_plotter
from .amber import write_amber_restraints

# Umbrella Integration
# from .umbrella_integrate import umbrella_integrate

# File conversion
from .convert import pdb2traj  # needs documentation
from .convert import gro2traj  # needs documentation
from .convert import xyz2traj  # needs documentation

# import subpackages

from . import styles
from . import units

# from . import dl_poly
from . import dna
from . import signal
from . import tunnel
from . import mutate
from . import monte
from . import rdkit

from .amino import alphabet as amino_alphabet
from .amino import longname as amino_longname

from . import hijack



"""FakeAtoms and FakeSurfaceCalculator to run ASE optimisations and dynamics within an N-dimensional PMF/energy surface"""

from .hijack import FakeAtoms, FakeSurfaceCalculator, NaNEncounteredInPmfError



import mout


class Atom:
    """Fundamental unit for molecular systems.

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    name_symbol_dict = {
        "MG": "Mg",
        "Mg": "Mg",
        "LIT": "Li",
        "Li": "Li",
        "SOD": "Na",
        "POT": "K",
    }

    def __init__(
        self,
        name,
        index=None,
        pdb_index=None,
        position=None,
        residue=None,
        chain=None,
        res_number=None,
        charge=None,
        FF_atomtype=None,
        mass=None,
        LJ_sigma=None,
        LJ_epsilon=None,
        occupancy=None,
        temp_factor=None,
        heterogen=None,
        charge_str=None,
        velocity=None,
        alternative_site=None,
        res_index=None,
        element=None,
        element_guess_warning=True,
    ):

        # necessary upon init
        self._name = name
        self.index = index
        if pdb_index is not None:
            self._NUMBER = int(pdb_index)
        else:
            self._NUMBER = None
        self._position = position
        self.residue = residue

        if element:
            self._element = element
        else:
            if element_guess_warning:
                mout.warning("Guessing element from first character of atom name!")
            self._element = name[0]
        assert self.symbol is not None

        self.chain = chain
        self.chain_number = None
        self._atomic_number = None
        if res_number is not None:
            self._res_number = int(res_number)
        else:
            self._res_number = None
        if res_index:
            self.res_index = int(res_index)
        else:
            self.res_index = None

        # optional
        self.FF_atomtype = FF_atomtype

        if charge:
            self.charge = charge
        elif charge_str:
            if charge_str.endswith("-"):
                self.charge = -1.0 * float(charge_str[:-1])
            else:
                self.charge = float(charge_str[:-1])
        else:
            self.charge = 0.0

        self._mass = mass
        self.LJ_sigma = LJ_sigma
        self.LJ_epsilon = LJ_epsilon
        self.occupancy = occupancy
        self.temp_factor = temp_factor
        self.heterogen = heterogen
        self.charge_str = charge_str
        self.ter_line = None
        self.terminal = None
        self._alternative_site = alternative_site
        self._velocity = velocity

        self._parent = None

    @property
    def element(self):
        return self._element

    @property
    def symbol(self):
        return self._element

    @property
    def species(self):
        return self._element

    @property
    def alternative_site(self):
        return self._alternative_site

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, obj):
        # import weakref
        # self._parent = weakref.ref(obj)
        self._parent = obj

    def __deepcopy__(self, memodict={}):
        copy_object = Atom(
            self.name,
            self.index,
            self.pdb_index,
            self.position,
            self.residue,
            res_number=self.res_number,
            res_index=self.res_index,
            element=self.element,
        )
        copy_object.chain = self.chain
        copy_object.occupancy = self.occupancy
        copy_object.temp_factor = self.temp_factor
        copy_object.heterogen = self.heterogen
        copy_object.charge_str = self.charge_str
        copy_object.charge = self.charge
        copy_object.velocity = self.velocity
        copy_object.chain_number = self.chain_number
        copy_object._alternative_site = self.alternative_site

        return copy_object

    def print(self):
        """Verbose output of the atom's properties"""
        import mout

        mout.varOut("Atom Name", self._name)
        mout.varOut("Atom Species", self.species)
        mout.varOut("Atom (ASE) Index", self.index)
        mout.varOut("Atom (PDB) Index", self.pdb_index)
        mout.varOut("Atom Position", self._position)
        mout.varOut("Atom Residue", self.residue)
        mout.varOut("Atom Chain", self.chain)
        mout.varOut("Atom Residue Number", self.res_number)

    def summary(self):
        """Summarised output of the atom's properties"""
        print(
            f"Atom {self.name} ({self.element}), index={self.index}, pdb_index={self.pdb_index}, res={self.residue}, res_number={self.res_number}, chain={self.chain}, chain_number={self.chain_number}"
        )

    def get_name(self, wRes=False, noPrime=False):
        """Returns a string of resname_atomname"""
        if wRes:
            namestring = self.residue + "_" + self._name
        else:
            namestring = self._name
        if noPrime:
            return namestring.replace("'", "P")
        else:
            return namestring

    def set_name(self, name, element, verbosity=1):
        """Rename the atom"""
        if verbosity > 0:
            import mout
            import mcol

            mout.out(
                "Renaming atom "
                + mcol.arg
                + self.name
                + str([self.index])
                + mcol.clear
                + " of residue "
                + mcol.varName
                + self.residue
                + str([self.res_number])
                + mcol.clear
                + " to "
                + mcol.arg
                + name
            )
        self._name = name

        assert element
        self._element = element
        assert self.symbol is not None

    @property
    def res_number(self):
        """number or pdb_index are the fixed residue index from the original structure file"""
        return self._res_number

    @property
    def number(self):
        """number or pdb_index are the fixed atom index from the original structure file"""
        return self._NUMBER

    @property
    def pdb_index(self):
        """number or pdb_index are the fixed atom index from the original structure file"""
        return self._NUMBER

    @property
    def name_number_str(self):
        return f"{self.name} {self.number}"

    @property
    def name(self):
        """Name (str) property"""
        return self.get_name()

    @name.setter
    def name(self, name):
        self.set_name(name=name, verbosity=0)

    @property
    def atomic_number(self):
        """Atomic number (int) property"""
        from ase.data import atomic_numbers as ase_atomic_numbers  # Atom only

        return ase_atomic_numbers[self.species]

    @property
    def covalent_radius(self):
        """Covalent radius (float) property"""
        from ase.data import covalent_radii as ase_covalent_radii  # Atom only

        return ase_covalent_radii[self.atomic_number]

    @property
    def vdw_radius(self):
        """vdW radius (float) property"""
        from ase.data import vdw_radii as ase_vdw_radii  # Atom only

        return ase_vdw_radii[self.atomic_number]

    @property
    def mass(self):
        """Atomic mass (float) property"""
        if self._mass is None:
            from ase.data import atomic_masses as ase_atomic_masses  # Atom only

            return ase_atomic_masses[self.atomic_number]
        else:
            return self._mass

    @property
    def position(self):
        """Cartesian position (list)"""
        return self._position

    @property
    def velocity(self):
        """Cartesian velocity (list)"""
        return self._velocity

    @property
    def np_pos(self):
        """Cartesian position (numpy array)"""
        import numpy as np

        return np.array((self._position[0], self._position[1], self._position[2]))

    @property
    def np_vel(self):
        """Cartesian velocity (numpy array)"""
        import numpy as np

        return np.array((self.velocity[0], self.velocity[1], self.velocity[2]))

    @property
    def x(self):
        """X-coordinate"""
        return self._position[0]

    @property
    def y(self):
        """Y-coordinate"""
        return self._position[1]

    @property
    def z(self):
        """Z-coordinate"""
        return self._position[2]

    @position.setter
    def position(self, pos):
        self._position = pos

    @velocity.setter
    def velocity(self, vel):
        self._velocity = vel

    def copy(self):
        """Returns a deepcopy of the Atom object"""
        import copy

        return copy.deepcopy(self)

    def is_in_residue(self, residue):
        from .residue import Residue

        assert isinstance(residue, Residue)

        # print(f"self: {self.residue}, {self.res_number}, {self.chain}")
        # print(f"other: {residue.name}, {residue.number}, {residue.chain}")

        if self.residue != residue.name:
            return False
        if self.res_number != residue.number:
            return False
        if self.chain != residue.chain:
            return False
        return True

    @property
    def type(self):
        from .residue import res_type

        return res_type(self.residue)

    @property
    def ase_atom(self):
        from ase import Atom

        return Atom(self.species, self.position)

    def __float__(self):
        return self.np_pos

    def __sub__(self, other):
        if isinstance(other, Atom):
            return self.np_pos - other.np_pos
        else:
            return self.np_pos - other

    def __add__(self, other):
        import numpy as np

        if isinstance(other, np.ndarray):
            return self.np_pos + other
        if isinstance(other, Atom):
            return self.np_pos + other.np_pos
        raise TypeError

    @property
    def children(self):
        return None

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    # def __del__(self):
    #   mout.debug(f'Atom({self},id({id(self)})) was deleted')
